## Import Libraries -- optimization based on pymoo
import numpy as np
from pymoo.core.problem import Problem
from pymoo.algorithms.moo.nsga3 import NSGA3
from pymoo.factory import get_reference_directions
from pymoo.optimize import minimize
from pymoo.util.termination.f_tol import MultiObjectiveSpaceToleranceTermination
import pandas as pd
from pymoo.core.callback import Callback

## Create an object for the problem formulation
class MyProblem(Problem):

    def __init__(self, A_ik, Lim_k, H_k, ET_cold, alpha_1, alpha_2, alpha_3, alpha_4, alpha_5, ET_warehouse,
                 FC_warehouse, beta_1, beta_2, beta_3, ET_tank, gamma, L_i, sigma_i, N_it, SW1, SW2, SW3,
                 cold_prods, tanks, warehouse_prods, **kwargs):

        # Base variables
        self.A_ik = A_ik
        self.Lim_k = Lim_k
        self.H_k = H_k
        self.n_t = N_it.shape[1]

        # Cold storage
        self.ET_cold = ET_cold
        self.alpha_1 = alpha_1
        self.alpha_2 = alpha_2
        self.alpha_3 = alpha_3
        self.alpha_4 = alpha_4
        self.alpha_5 = alpha_5

        # Warehouse
        self.ET_warehouse = ET_warehouse
        self.FC_warehouse = FC_warehouse
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.beta_3 = beta_3

        # Tanks
        self.ET_tank = ET_tank
        self.gamma = gamma

        # Shortage
        self.L_i = L_i
        self.sigma_i = sigma_i
        self.N_it = N_it
        self.SW1 = SW1
        self.SW2 = SW2
        self.SW3 = SW3

        ## Support math
        self.S_cold_support = np.matmul(self.A_ik, self.ET_cold)
        self.S_ware_support = np.matmul(self.A_ik, self.ET_warehouse)

        ## Boundaries
        xl = np.zeros(self.A_ik.shape[0])

        # xu = np.zeros(A_ik.shape[0])
        xu = np.zeros(self.A_ik.shape[0])
        xu[0:cold_prods] = Lim_k[0]
        xu[-warehouse_prods:] = Lim_k[-1]
        for i in range(tanks):
            xu[cold_prods + i] = Lim_k[i + 1]

        super().__init__(n_var=self.A_ik.shape[0],
                         n_obj=3,
                         n_constr=self.A_ik.shape[1],
                         xl=xl,
                         xu=xu,
                         **kwargs)

    def _evaluate(self, x, out, *args, **kwargs):

        pop_size = x.shape[0]

        ## Calculate Energy Consumption
        # Cold Storage Energy Consumption
        S_cold = np.matmul(x, self.S_cold_support)

        E_cold = (self.alpha_2 /
                  (1 + np.exp(-((S_cold - 1) / self.alpha_1 - np.floor((S_cold - 1) / self.alpha_1)) *
                              self.alpha_5 / self.alpha_4 + self.alpha_5)
                   ) + np.ceil(S_cold / self.alpha_1)
                  ) * self.alpha_3

        # Warehouse Energy Consumption
        S_warehouse = np.matmul(x, self.S_ware_support)
        E_warehouse = self.FC_warehouse + S_warehouse * self.beta_1 + self.beta_2 - S_warehouse * self.beta_3

        # Tanks Energy Consumption
        S_k = np.matmul(x, self.A_ik)
        E_tank_temp = S_k / self.Lim_k
        E_tank_temp = E_tank_temp[:, 1:-1]
        E_tank = np.matmul(E_tank_temp, self.gamma.reshape(-1, 1))

        # Total Energy Consumption
        E = E_cold + E_warehouse + E_tank

        ## Calculate Holding Costs
        C_hold = np.matmul(S_k / self.Lim_k, self.H_k.reshape(-1, 1))

        ## Calculate Shortage Risk
        B_it = np.zeros([pop_size, self.n_var, self.n_t])
        B_it[:, :, 0] = x - self.N_it[:, 0]
        for i in range(1, self.n_t):
            B_it[:, :, i] = B_it[:, :, i - 1] - self.N_it[:, i]

        B_it_sign = ((np.sign(B_it) + 1) > 0).astype(int) * 2 - 1
        ShD_i = (self.n_t - (self.n_t / 2 - np.sum(B_it_sign, axis=2) / 2 - 1))

        Sh1 = (self.L_i - np.transpose(ShD_i)) * ((self.L_i - np.transpose(ShD_i)) > 0).astype(int)
        Sh2 = (self.sigma_i + self.L_i - np.transpose(ShD_i)) * (
                    (self.sigma_i + self.L_i - np.transpose(ShD_i)) > 0).astype(int) * (
                      (Sh1 == 0) + 0)
        Sh3 = (-((-(self.L_i + self.sigma_i - np.transpose(ShD_i))) * (
                    (-(self.L_i + self.sigma_i - np.transpose(ShD_i))) > 0).astype(
            int)) + self.n_t) * \
              ((((Sh1 > 0) | (Sh2 > 0) | (np.transpose(ShD_i) == self.n_t + 1)) == 0) + 0)

        R_sh = np.sum(Sh1 * self.SW1, axis=0) + np.sum(Sh2 * self.SW2, axis=0) + np.sum(Sh3 * self.SW3, axis=0)

        # Finalize Objective Values
        objs = np.column_stack([E, C_hold, R_sh])

        ## Calculate constraints
        constr = (S_k - self.Lim_k)

        ## Store some variables
        self.Sh1 = Sh1
        self.Sh2 = Sh2
        self.Sh3 = Sh3
        self.R_sh2 = R_sh
        self.S_k = S_k

        out["F"] = objs
        out["G"] = constr


## Create a callback function to provide additional metrics during the optimization
class MyCallback(Callback):

    def __init__(self) -> None:
        super().__init__()
        self.data["best"] = []
        self.data["mean"] = []

    def notify(self, algorithm):
        self.data["best"].append(algorithm.pop.get("F").min())
        self.data["mean"].append(algorithm.pop.get("F").mean())


## Values used
cold_prods = 50
tanks = 24
warehouse_prods = 250

n_i = cold_prods + tanks + warehouse_prods
n_t = 60

# Cold storage
ET_cold = np.zeros([tanks + 2, 1])
ET_cold[0, 0] = 1
alpha_1 = 50000
alpha_2 = 0.4
alpha_3 = 135
alpha_4 = 0.7
alpha_5 = 8

# Warehouse
ET_warehouse = np.zeros([tanks + 2, 1])
ET_warehouse[-1, 0] = 1
FC_warehouse = 100
beta_1 = 4e-4
beta_2 = 250
beta_3 = 3e-5

# Tanks
ET_tank = np.zeros([tanks + 2, tanks])
for i in range(tanks):
    ET_tank[i + 1, i] = 1

SW1 = 100
SW2 = 10
SW3 = 0.1

## Import datasets
A_ik = pd.read_csv("data/A_ik.csv", header=None).to_numpy()
Lim_k = pd.read_csv("data/Lim_k.csv", header=None).to_numpy()[:, 0]
H_k = pd.read_csv("data/H_k.csv", header=None).to_numpy()[:, 0]
gamma = pd.read_csv("data/gamma.csv", header=None).to_numpy()[:, 0]
L_i = pd.read_csv("data/L_i.csv", header=None).to_numpy()
sigma_i = pd.read_csv("data/sigma_i.csv", header=None).to_numpy()
N_it = pd.read_csv("data/N_it.csv", header=None).to_numpy()


# Define the non-parallelized problem
ref_dirs = get_reference_directions("das-dennis", 3, n_partitions=12)

# Create the algorithm object
pop_size = 200
max_gen = 2000
algorithm = NSGA3(pop_size=pop_size,
                  ref_dirs=ref_dirs)

termination = MultiObjectiveSpaceToleranceTermination(
    tol=0.1,
    # tol=0.0025,
    n_last=30,
    nth_gen=10,
    n_max_gen=max_gen,
    n_max_evals=None
)

current_max_version = 0
desired_version = 10
temps_desired = 30
save_folder = 'results_temporal_evolution2'

## Iterates over a temporal range
for iter_version in range(current_max_version+1, desired_version+1):

    print('VERSION:', iter_version)
    for temporal_iter in range(temps_desired):

        # Subset the production schedule
        print('Temporal Range: ', temporal_iter + 1)
        N_it_iter = N_it[:,(0 + temporal_iter):(60 + temporal_iter)]

        problem_test1 = MyProblem(A_ik, Lim_k, H_k, ET_cold, alpha_1, alpha_2, alpha_3, alpha_4, alpha_5, ET_warehouse,
                                  FC_warehouse, beta_1, beta_2, beta_3, ET_tank, gamma, L_i, sigma_i, N_it_iter, SW1, SW2, SW3,
                                  cold_prods, tanks, warehouse_prods)

        res = minimize(
            problem_test1,
            algorithm,
            termination,
            callback=MyCallback(),
            save_history=True,
            verbose=False
        )

        ## Store the results
        txt_file = str(temporal_iter + 1) + '_results'
        with open('results_temporal_evolution/' + txt_file + "_v" + str(iter_version) + ".txt", 'w') as text_file:
            text_file.write(
                'Start Time: ' + str(res.start_time) +
                '\nEnd Time: ' + str(res.end_time) +
                '\nExecution Time: ' + str(res.exec_time) +
                '\nBase Population Size: ' + str(res.pop.size) +
                '\nFinal Population Size: ' + str(res.opt.size) +
                '\nMax Generations: ' + str(max_gen) +
                '\nRan Iterations: ' + str(len(res.history)) +
                '\nTemporal Iteration: ' + str(temporal_iter + 1)
            )

        pd.DataFrame(res.F).to_csv(
            save_folder + "/" + str(temporal_iter + 1) + "_F_v" + str(iter_version) + ".csv", header=False,
            index=False)
        pd.DataFrame(res.G).to_csv(
            save_folder + "/" + str(temporal_iter + 1) + "_G_v" + str(iter_version) + ".csv", header=False,
            index=False)
        pd.DataFrame(res.X).to_csv(
            save_folder + "/" + str(temporal_iter + 1) + "_X_v" + str(iter_version) + ".csv", header=False,
            index=False)
        pd.DataFrame(res.algorithm.callback.data["best"]).to_csv(
            save_folder + "/" + str(temporal_iter + 1) + "_losses_best_v" + str(iter_version) + ".csv",
            header=False, index=False)
        pd.DataFrame(res.algorithm.callback.data["mean"]).to_csv(
            save_folder + "/" + str(temporal_iter + 1) + "_losses_mean_v" + str(iter_version) + ".csv",
            header=False, index=False)